package WLib;
use strict;
use Carp;
require Exporter;
use vars qw( @ISA @EXPORT @EXPORT_OK $debug $os $onLine );
use vars qw( %dnsCache $myIpStr $dnsInstalled $dnsTimeout 
  $running $exit $dontWait $network $cachingDns %authentication); 
use vars qw( $socketError );

@ISA = qw(Exporter);
@EXPORT_OK = qw($debug $os $onLine $dnsInstalled $dnsTimeout 
            $myIpStr $dontWait mergeTrees resolveIP myIP 
            nonBlockingConnect getOS isInSubnet 
            isInSubnets isOnLine isConnectable 
            refreshIp unarchive stringify $socketError getNameServers 
            $network firewall getXML putXML parseNetworkInfo getProxies 
            $cachingDns initAuthentication getCredentials 
            updateCredentials showCredentials getURL $running $exit 
            killSock);

#Windows systems have a limitation on the number of sockets 
#we can create at a time. Some kind of limit on memory allocated, 
#I think. Unfortunately it's unpredictable, and no error is 
#returned when this happens. It can be said that older versions 
#have a lower limit though. We dynamically adjust the following 
#default by using a 'canary' socket (see below in isConnectable).
our $maxSocketsAllowed = $^O =~ /MSWin/ ? 60 : 300;
$debug = (defined($debug) ? $debug : 1);
$os = -1;
$onLine = undef; #must be undef, or onLine test 
                #doesn't test whole layer 0 at startup
$myIpStr = '';
%dnsCache = (); $cachingDns = 1; $dnsCache{localhost} = '127.0.0.1';
$dontWait = 1;
$socketError = 0;
$dnsInstalled = 1;
my %network = ();
$network = \%network;

=head1 NAME

WLib - some useful routines for harassing the censors

=head1 SYNOPSIS

use WLib qw(mergeTrees);
$newTree = mergeTrees($oldTree, $newTree[, $rootPathName, $rootDepth]);
etc.

=head1 DESCRIPTION

=head1 METHODS

The following methods are available:

=over 4

=cut

use vars qw(@ISA $VERSION);

$VERSION = sprintf("%d.%02d", q$Revision: 4.156 $ =~ /(\d+)\.(\d+)/);
if ($debug>1) {print "Loaded WLib.pm version: $WLib::VERSION\n"};

BEGIN {
  if (eval {require Compress::Zlib}) {import Compress::Zlib}
  else {unless (defined(&Z_OK)) {eval ('sub Z_OK {}')}};
  if (eval {require Archive::Zip}) {import Archive::Zip};
  };
  
#try to catch errors at lp build and wait for user to see.
#don't wait if we were running already (so user can 
#'stop' and 'start' normally.
#don't wait in WLib if $exit is set
END {
  unless ($running or $exit or $dontWait) {
    print "Press Enter to close this window"; <STDIN>}
    }; 

sub getProxies { #get any known proxies we can find.
  #XXX - in mozilla/Netscape/preferences etc. too? 
  #XXX - Look for Program Files/Mozilla.org, Netscape etc.
  #XXX - for auto, we can't use any firewall httpProxyFqdn info
  #XXX - but we could get this info as well for other ISPs.
  my ($proxies, %proxies, $envProxy, $regProxies, @regProxies) = ();
  for my $env (keys(%ENV)) { #must search the whole env because there 
                             #is no case standard for these strings
    if ($env =~ /^http_proxy$/i) { #case insensitive match
      $ENV{$env} =~ m|(http://)?(.*:\d+)|;
      (undef, $envProxy) = ($1, $2);
      unless ($envProxy) {$envProxy = ''};
      $proxies{$envProxy} = ();
      };
    };
    
  if ($debug) {
    print "Proxies found in environment: ", 
    %proxies ? join(', ', keys(%proxies)) : 'none', "\n"
    };
    
  if ($^O =~ /MSWin32/ ) { #see if we can get anything useful from the 
    #registry
    #Only good if the user hasn't already overwritten it with the 
    #localProxy settings
    require Win32::TieRegistry;
    my $connsKey = $Win32::TieRegistry::Registry->
      {"HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\".
      "CurrentVersion\\Internet Settings\\Connections"};
    for my $conn (keys(%{$connsKey})) { #*value* for each connection
      next unless $connsKey->{"$conn"};
      $regProxies .= ', '.stringify($connsKey->{"$conn"});
      if ($debug>3) {print $conn, stringify($connsKey->{"$conn"}), "\n"};
      };
    unless ($regProxies) {$regProxies = ''};
    @regProxies = ($regProxies =~ m|https?=([^;\0:]*:\d+);|ig);
    };
  if ($debug>3) {print "regProxies: ", join(', ', @regProxies), "\n"};
  #$proxies = join(', ', $envProxy, @regProxies);
  #@proxies{split(/,\s*/, $proxies)} = (); #get rid of dupes
  @proxies{@regProxies} = (); #remove dupes
  
  #Find any in firewalls.xml ... assume no firewall if none known
  my $fwR = getXML('firewalls.xml');
  my @firewalls = (firewall(myIP(), $fwR) or 'firewall-none');
  my @fwProxies = ();
  for my $fw (@firewalls) { 
    if ($fwR->{$fw}->{httpProxyFqdn}) {
      my @fwBits = split(/,\s*/, $fwR->{$fw}->{httpProxyFqdn});
      for my $fwBit (@fwBits) {
        #my ($fwBitAddr, $fwBitPort) = split(/:/, $fwBit);
        #my @fwBitIPs = ();
        #if ($fwBitAddr =~ /^\d+\.\d+\.\d+\.\d+$/) {
          #$fwBitIPs[0] = $fwBitAddr;
          #}
        #else {
          #@fwBitIPs = resolveIP($fwBitAddr);
          #};
        #for my $ip (@fwBitIPs) {
          #push(@fwProxies, "$ip:$fwBitPort");
        push(@fwProxies, $fwBit);
          #};
        };
      };
    };
  if ($debug > 3) {
    print "Firewalls.xml proxies: ", join(', ', @fwProxies), "\n"
    };
  @proxies{@fwProxies} = ();
  
  $proxies = join(', ', keys(%proxies));
  if ($debug>2) {print "Known proxies: $proxies\n"};
  return wantarray ? keys(%proxies) : $proxies;
  };

=item initAuthentication($config->{authentication}, 
  $commandLineAuthStr, $userConfig)
Initialize all the authentication info we can find. 
Taken from environment, command line, and user configuration file

=cut

sub initAuthentication { #get auth info
  my ($authR, $auth, $userConfig) = @_;
  #Store the env realm and defaulted realms (realm left out in the 
  #command line auth spec) as realms realm_realm0, realm_realm1, etc. 
  #The env may supply only one realm. If $ENV{HTTP_proxy_realm} 
  #is defined (the user has obtained this from a browser popup window, 
  #for example), then we use that. Otherwise the env realm defaults to 
  #realm_realm0, and will never match requests, so auth info will only 
  #be sent by a proxyTools app when the service matches. 
  #Clients (news etc.) will do their own thing as normal, failing 
  #when a firewall asks for auth.
  #When a request is received, LP should intervene (if it has matching 
  #auth info) under all circumstances except possibly when the user is 
  #using a web browser (HTTP protocol). This would allow proxyTools and 
  #non-web apps to use a CONNECT out through a firewall/proxy, for 
  #example). 

  #XXX - A web client (user) response could be watched and the 
  #XXX - auth info and realm obtained by LP for Basic authentication. 
  #XXX - The idea fails for all other auth methods where replay 
  #XXX - attacks (which this would be) are specifically guarded against.
  #XXX - It also fails where it's expected that the authentication 
  #XXX - process gives the browser a cookie whic is used to authenticate 
  #XXX - the rest of the site (though this *could* probably be handled by 
  #XXX - sending the credentials preemptively with each subsequent request.

  #XXX - LP2 sees 401 and 407 (and firewall, 501?) responses, but doesn't 
  #XXX - intervene unless it has the required auth info. In the 501 case, 
  #XXX - the status code may not exist, and the the realm will not be in 
  #XXX - the response.
  #XXX - The user may recognize from 
  #XXX - his client's response that authentication was required. We 
  #XXX - provide a GUI button to allow him to query LP2 to see the last 
  #XXX - realm and service(s?) for which authentication was requested, 
  #XXX - and then to add his (cleartext) username and password to LP2's 
  #XXX - auth data. 
  #XXX - LP2 will then be able to handle his next attempt.
  #XXX - This way, the GUI knows the answers and can update the user's 
  #XXX - config too, so the user need not do this again.
  
  #XXX - Will I be able to guarantee to always be able to recognize 
  #XXX - the incoming auth request properly though? It might end up 
  #XXX - being like recognizing access-denied web page resonses.
  #XXX - I'm guessing that firewalls send the request, to any 
  #XXX - through-firewall connection attempt, in a 
  #XXX - form that a web browser understands (it looks like it comes 
  #XXX - from a server or proxy). Any other format, I won't recognize 
  #XXX - (until someone sends me a data fragment).
  
  #XXX - LP2 doesn't know the service is actually requesting 
  #XXX - the authentication (it might be the firewall), 
  #XXX - unless it already has the firewall realm. How does it know 
  #XXX - it has this? Once it knows it has this, it can assume any other 
  #XXX - auth requests are from the service we're trying to use. Then we 
  #XXX - know the service. OTOH, as long as the user enters his password 
  #XXX - for a realm, LP2 can use it for that realm any time without 
  #XXX - knowing the service (and note that the service might have been 
  #XXX - spec'd as an fqdn like proxy1.emirates.net.ae:8080, so we can't 
  #XXX - easily check against an IP address anyway); 
  #XXX - it just means a resend each time until LP2 
  #XXX - is sure this is the firewall. How will it become sure? 
  #XXX - Note login.icq.com, or 
  #XXX - www.panix.com never ask for auth, so LP2 could make an 
  #XXX - initial test to see if there was a firewall...
  #XXX - Even for commStrat 1, the service requesting auth will 
  #XXX - always be the last layer accessed (once we are sure the request 
  #XXX - is not from a/the firewall). 
  
  #XXX - LP should resend most requests though (and to the same 
  #XXX - service!), if it has the auth info available, and knows 
  #XXX - the service. 

  #XXX - Always print the realm from any requests received, so the 
  #XXX - user can save it in his env, config, or use it on a command line.

  #If the user specifies auth info on the command line, we need 
  #[realm],service,username,password. Only *one* of realm, or service 
  #may be omitted.
  #For firewall auth there is no proxy:port, so the realm must be given.

  #Note that we may need auth info per protocol (at least where only the 
  #realm has been specified). This will probably never happen.

  #Note that some firewalls insert a random number (a nonce) in the realm 
  #name (or following it, I guess) every time a user authenticates. 
  #We might need to allow a 'soft match'
  #when checking if we have the auth info for such realms (scan through 
  #all the auth info and regex match, I guess).
  #This should not be necessary if we know the realm field precisely.
  #Not sure how this will pan out.
  
  #Note that the protocol on any environment spec'd proxy seems 
  #redundant, given that the env var name itself also contains this 
  #info. We strip that off here...
  #We don't handle basic authentication for web site URIs in LP; the 
  #browser does that. So there's no need to store info for the basic 
  #authentication URIs at all. Proxy basic auth is always from the 
  #doc root and we have that by storing auth info for service_<address>
  #Firewall auth can have no address,but we do that by storing 
  #realm_<firewallRealm>.

  #Store all info as both 
  #$authentication{realm_<realmName>} =  
  #$authentication{service_<service>} = 
  #realm,service,username,password
  #to allow easy access, where <service> is likely to be <proxyAddress:Port>, 
  #but may also be things like <webSiteAddress:Port> or 
  #<newsServerAddress:Port> etc. 
  #Each service can only need auth info for one protocol
  #Each realm - dunno, but I'm assuming only one auth string

  my $realmNr = 0; #environment before anything else
  my $realm = "realm$realmNr";
  my $realmSoft;
  my %pa = ();
  #do a complete search since there seem to be all sorts 
  #of variation in case
  for my $env (keys(%ENV)) {
    for my $match ('HTTP_proxy', 'HTTP_proxy_user', 'HTTP_proxy_pass', 
      'HTTP_proxy_realm') {
      next unless $env =~ /^$match$/i;
      $pa{$match} = ($ENV{$env} =~ m|^(http://)?(.*)/?$|)[1];
      };
    };
  if ($pa{HTTP_proxy_realm}) {$realm = $pa{HTTP_proxy_realm}};
  if ($pa{HTTP_proxy} and $pa{HTTP_proxy_user} and $pa{HTTP_proxy_pass}) {
    my ($addr, $port) = split(/:/, $pa{HTTP_proxy});
    $authR->{"service_$pa{HTTP_proxy}"} = 
    $authR->{"realm_$realm"} = 
      "$realm,$pa{HTTP_proxy},$pa{HTTP_proxy_user},$pa{HTTP_proxy_pass}";
    #also store 'soft' realm (all digits replaced by x) for possible match
    if (($realmSoft = $realm) =~ s/\d/x/g) {
      $authR->{"realm_$realmSoft"} = 
        "$realmSoft,$pa{HTTP_proxy},$pa{HTTP_proxy_user},$pa{HTTP_proxy_pass}";
      };    
    unless ($pa{HTTP_proxy} =~ /\d+\.\d+\.\d+\.\d+/) { #fqdn
      for my $ipAddr (resolveIP($addr)) { #auth ip addresses too!
        $authR->{"service_$ipAddr:$port"} = 
          "$realm,$ipAddr:$port,$pa{HTTP_proxy_user},$pa{HTTP_proxy_pass}";
        };
      };
    }
  else {
    print "No (or incomplete) credentials found in environment\n" 
      if $debug>2;
    };
    
  #Try user config - overwrites environment info
  if ($userConfig and $userConfig->{authentication}) { 
    mergeTrees($userConfig->{authentication}, $main::config->{authentication});
    };

  #and the command line...
  if (defined($auth)) {
    if ($auth !~ /^env$/i) {
      return addCredential($authR, $auth);
      }
    else { #user said '-a env'
      unless ($authR->{realm_realm0}) { #no env auth 
        print "No authentication info found in the environment\n";
        print "You need HTTP_proxy, HTTP_proxy_user and \n";
        print "HTTP_proxy_pass defined\n";
        return 0;
        };
      #use what we already have - the user has just confirmed that.
      #may remove this option value in future.
      }
    }
  else { #no '-a' flag at all
    #we may be running from LP (statProxy/autoconfig)
    #use what we have from env. We probably have the wrong realm  
    #for his firewall, so auth info will only get sent if user is 
    #testing a proxy specified in environment. In normal use, the 
    #client handles auth for services (web browsers will handle 
    #'firewall' authentication which probably really comes from 
    #irc clients etc. to a proxy), but we have things like  
    #statProxy, news, ftp, to consider. 
    #To test anything else, we *must* know the firewall realm.
    
    #XXX - We don't send auth info *anywhere* unless we 
    #XXX - match it up with a realm, or a service, so attempting to 
    #XXX - connnect to a remote service via a firewall which requires 
    #XXX - authentication will fail unless the user provides the auth 
    #XXX - info.
    };
  return 1;
  };
  
=item $userPass = getCredentials($config->{authentication}, $realm, 
  $dest0, $dest1);
  
Get credentials for use with the realm, service at $dest.

=cut

sub getCredentials { #get username and password appropriate for $dest
  my ($configA, $realm, $svc) = @_;
  my ($authStr, $realmSoft);
  #try hard to match the realm; anything else is a risk
  #order: realm exact match, realm soft match, service match. 
  if ($debug>-1) {
    print "Seeking credentials for ", 
      $realm ? $realm : '(no realm)', ", $svc\n";
    };
  if ($realm) {
    $authStr = $configA->{"realm_$realm"} or 
    ((($realmSoft = $realm) =~ s/\d/x/g) and 
     ($authStr = $configA->{"realm_$realmSoft"}))
     };
  #XXX - resolve service first, if necessary(?).
  if (!$realm and $svc) { #don't allow a null service
    $authStr = $configA->{"service_$svc"}
    };
            
  if ($authStr) { #got something
    if ($debug>-1) {print "Found auth string: $authStr\n"};
    my ($realm, $service, $user, $pass)= split(/\,/, $authStr); 
    return "$user:$pass";
    }
  else {return undef};
  };
  
sub updateCredentials { #set of auth strings sent from GUI
  #in format "authServer, authType, realm, service, user, pass", 
  #';' separated
  my $authStr = shift;
  if ($debug) {print "Adding credentials: $authStr\n"};
  my @auths = split(/;\s*/, $authStr);
  my $nrAuthsOk = 0; 
  for my $auth (@auths) {
    next unless $auth; #catch empty str from GUI blank lines etc.
    my ($authServer, $authType, $realm, $service, $user, $pass) = 
      split(/,\s*/, $auth);
    if ($pass) { #good string
      addCredential($main::config->{authentication}, 
        "$realm, $service, $user, $pass") and $nrAuthsOk++;
      }
    else {
      print "Bad credentials from user - ignored\n";
      }
    };
  delete $main::config->{authentication}->{requested};
  return $nrAuthsOk;
  }
  
sub addCredential { #add a credential to our database
  #format "realm, service, user, pass"  
  my ($authR, $auth) = @_;
  if ($debug) {print "Adding credential: $auth\n"};
  my $realmSoft;
  unless ($auth =~ /,.*,.*,/) {
    print "Bad authentication string: $auth\n";
    print "Must be ,<serviceAddress:port>,<username>,<password>\n";
    print "or <realm>,,<username>,<password>\n";
    return 0;
    };
  #he's specifying info for a new realm
  my ($realm, $service, $username, $pass) = 
    split(/,\s?/, $auth);
  unless ($username and $pass) { 
    print "No username or password\n";
    exit;
    };
  #allow user to override any environment settings if he specifies realm0
  if ($realm =~ /^realm0$/) {print "Discarding environment credentials\n"}
  if ($authR->{"realm_$realm"} or 
    $authR->{"service_$service"}) {
    print "Overwriting credentials for $realm,$service\n";
    }
  my $entry = join(',', ($realm, $service, $username, $pass));
  #only add entries if we actually have something
  if ($realm) {$authR->{"realm_$realm"} = $entry};
  if ($service) {$authR->{"service_$service"} = $entry};
  #add one for the resolved name too ..
  if ($service) {
    my ($addr, $port) = split(/:/, $service);
    if (($addr !~ /^\d+\.\d+\.\d+\.\d+$/) and $port) {
      $addr = resolveIP($addr);
      if ($addr) {$authR->{"service_$addr:$port"} = $entry};
      };
    };
  if (($realmSoft = $realm) =~ s/\d/x/g) {
    $authR->{"realm_$realmSoft"} = 
      join(',', ($realmSoft, $service, $username, $pass)); 
    };
  return 1;
  };  

#snatched from LWP::UserAgent::Authen::Digest
#Thanks GAAS!
#XXX - changed all the request object methods to hash refs (!!!)
sub authenticate {
  my($class, $ua, $proxy, $auth_param, $response,
     $request, $arg, $size, $user, $pass) = @_;
  #my($user, $pass) = $ua->get_basic_credentials($auth_param->{realm},
                                                #$request->url, $proxy);
  #return $response unless defined $user and defined $pass;
  my $nc = sprintf "%08X", #just a count
    ++$ua->{authen_md5_nonce_count}{$auth_param->{nonce}};
  my $cnonce = sprintf "%8x", time;
  my $uri = $request->{url}->{path_query};
  $uri = "/" unless length $uri;
  my $md5 = Digest::MD5->new;
  my(@digest);
  $md5->add(join(":", $user, $auth_param->{realm}, $pass));
  push(@digest, $md5->hexdigest);
  $md5->reset;

  push(@digest, $auth_param->{nonce});
  if ($auth_param->{qop}) {
    push(@digest, $nc, $cnonce, $auth_param->{qop});
    }
  $md5->add(join(":", $request->{method}, $uri));
  push(@digest, $md5->hexdigest);
  $md5->reset;

  $md5->add(join(":", @digest));
  my($digest) = $md5->hexdigest;
  $md5->reset;

  my %resp = map { $_ => $auth_param->{$_} } qw(realm nonce opaque);
  @resp{qw(username uri response algorithm)} = ($user, $uri, $digest, "MD5");
  if (($auth_param->{qop} || "") eq "auth") {
    @resp{qw(qop cnonce nc)} = ("auth", $cnonce, $nc);
    }

  my(@order) = qw(username realm qop algorithm uri nonce nc cnonce response);
  if($request->{method} =~ /^(?:POST|PUT)$/) {
    $md5->add($request->{content});
    my $content = $md5->hexdigest;
    $md5->reset;
    $md5->add(join(":", @digest[0..1], $content));
    $md5->reset;
    $resp{"message-digest"} = $md5->hexdigest;
    push(@order, "message-digest");
    }
  push(@order, "opaque");
  my @pairs;
  for (@order) {
    next unless defined $resp{$_};
    push(@pairs, "$_=" . qq("$resp{$_}"));
    }

  my $auth_header = $proxy ? "Proxy-Authorization" : "Authorization";
  my $auth_value  = "Digest " . join(", ", @pairs);

=for wayne

  # Need to check this isn't a repeated fail!
  my $r = $response;
  while ($r) {
    my $u = $r->request->{digest_user_pass};
    if ($u && $u->[0] eq $user && $u->[1] eq $pass) {
      # here we know this failed before
      $response->header("Client-Warning" =>
            "Credentials for '$user' failed before");
      return $response;
      }
    $r = $r->previous;
    }

  my $referral = $request->clone;
  $referral->header($auth_header => $auth_value);
   we shouldn't really do this, but...
  $referral->{digest_user_pass} = [$user, $pass];

  return $ua->request($referral, $arg, $size, $response);

=cut

  return $auth_value;
};


=item $tree = mergeTrees($fromTree, $toTree[, $rootName, $rootDepth]);
mergeTrees - a useful hash tree merge. Copy one hash tree to another, 
instantiating nodes and overwriting as needed.

=cut

sub mergeTrees { 
  #merge using pre-order tree traversal
  #traverse the first tree, merge *everything* to the 
  #second, overwrite/create as necessary
  my ($n, $o, $path, $depth) = @_;
  $path = '' unless defined $path;
  $depth = 0 unless defined $depth;
  #print "  " x $depth, "$depth\n";
  #first time we encounter a difference between trees, 
  #we instantiate the node in the dest tree.
  while (my ($sk, $sv) = each(%{$n})) {
    #if (defined($o->{$sk})) {}; #just instantiate?
    my $p = $path.'->{'.$sk.'}';
    if (ref($sv) and ref($o->{$sk})) { 
      mergeTrees($sv, $o->{$sk}, $p, $depth + 1);
      }
    elsif (ref($sv)) { #overwrite the dest value with a hash pointer!
      #print "overloading a value (", defined($o->{$sk}) ? $o->{$sk} 
        #: 'undef', ") with a pointer at $p\n";
      if (defined($o->{$sk})) {
         print "Overloading a value (", $o->{$sk}, 
          ") with a pointer at $p\n" if ($debug > 3);
        };
      $o->{$sk} = $sv; #now dest node points to source subtree
      }
    elsif (ref($o->{$sk})) { #overwrite a pointer with a value (scary!)
      print 
        "warning: Overloading a whole subtree with a value ($sv) at $p\n" 
        if ($debug > 0);
      $o->{$sk} = $sv;
      }
    else { #overwriting a value with a new value
      if (defined($o->{$sk}) and ($o->{$sk} ne $sv)) {
        print "Overloading $p (was: $o->{$sk}, now: $sv)\n" if 
          ($debug > 3);
        };
      $o->{$sk} = $sv;        
      };
    #if ($path=~/allLayer/) {print "sk, sv: $sk, $sv\n"};
    };
  return $o;
  }
  
=item $IPAddress = resolveIP($fqdn, $timeout, $dnsList);
resolveIP - a non-blocking DNS resolution routine.

=cut

sub resolveIP { #do nonblocking fqdn->IP address resolution for this host
  my ($fqdn, $timeout, $nsList) = @_;
  if ($dnsCache{$fqdn} and $cachingDns) {
    if ($dnsCache{$fqdn} =~ /\./) { #normal entry (IP address)
      if ($debug>4) {print "DNS cache hit: $fqdn\n"};
      return wantarray ? split(/, /, $dnsCache{$fqdn}) : $dnsCache{$fqdn};
      }
    else { #no dot, so it's a failure count
      return undef if ($dnsCache{$fqdn} > 2); #avoid timeouts when dns is failing
      }
    }
  else { #not cached
    my @results;
    $timeout = 5 unless $timeout; #default 5
    my $res = new Net::DNS::Resolver;
    
    #XXX - experimenting with these
    #$res->retrans(3); # retransmission time (default is 5)
    #$res->retry(3); # nr of retries (default is 4)
    #$res->nameservers() on Windows normally contains '127.0.0.1' :-(
    my @nameServers;
    if ($nsList) {
      @nameServers = split(/, /, $nsList)
      }
    elsif ($network->{provider}->{nameServer}) {
      @nameServers = 
        split(/, /, $network->{provider}->{nameServer})
      }
    elsif (@nameServers = split(/, /, getNameServers())) {}
    #else {$res->nameservers('192.5.6.30')}; #start at the top!;
    #else {@nameServers = ('152.163.159.234')}; #start icq
    else {@nameServers = ('192.112.36.4')}; #start at G.ROOT-SERVERS.NET
    $res->nameservers(@nameServers);
    if ($debug>3) {
      print "Using name servers (timeout: $timeout): ", 
        join(', ', $res->nameservers), "\n";
      };

    #XXX - vc's might be good here - need reliability, not speed.
    #and I have a nonblocking tcp connect now ... do all ISP's nameservers 
    #(and firewalls) allow access to it though?
    
    #Note bgsend does not do the automatic failover to other name 
    #servers that query would normally do. Need to do that here.
    my $nsIndex = 0;
    while ($nsIndex < @nameServers) { 
      my $skt = $res->bgsend($fqdn); #will do PTR if it needs to  
      my $sel = new IO::Select($skt);
      if ($sel->can_read($timeout)) {
        #we can only have $skt here
        my $packet = $res->bgread($skt);
        if (($debug>5) and $packet) {$packet->print};
        if ($packet) { #XXX - we can detect that we are off-line.
          foreach my $rr ($packet->answer) {
            my $type=$rr->type;
            if (($type eq "A") and ($rr->address)) {
              if ($debug>3) {
                print "Resolved $fqdn to ", $rr->address, " (via DNS)\n"
                };
              push(@results, $rr->address);
              last unless wantarray;
              }
            };
          if (wantarray) {
            $dnsCache{$fqdn} = join(', ', @results);
            return @results
            }
          else {
            $dnsCache{$fqdn} = $results[0];
            return $results[0]
            };
          }
        }
      #failed, try next name server (rotate array left)
      $nsIndex++;
      if ($debug>2) {print "Rotating name servers: $nsIndex\n"};
      push(@nameServers, shift(@nameServers));
      $res->nameservers(@nameServers);
      }
    if ($debug) {print "\nDns (", join(', ', $res->nameservers), 
      ") timeout ($timeout) resolving $fqdn - check if online\n"};
    $dnsCache{$fqdn}++; #failure count
    return undef;
    };
  };

=item $connected = $nonBlockingConnect($hostPort, $timeout);
NonBlocking TCP connect, times out after $timeout if no connection.
This method is of most use in MS Windows where forking and blocking 
are not properly implemented. Timeout works now! 
Returns the socket handle (or undef on timeout).

=cut

BEGIN {
   # This nonsense is needed in 5.6.1 and earlier
   if( $^O eq 'MSWin32' ) {
      *EWOULDBLOCK = sub () { 10035 }; 
      *EAGAIN = sub () { 10035 }; 
      *EINPROGRESS = sub () { 10036 };
      *ENOBUFS =  sub () { 10055 }; 
      *ECONNREFUSED = sub () { 10061 };
      *EMFILE = sub () { 10024 };
      *EINVAL = sub () { 10022 };
      *IO::Socket::blocking = sub {
          my ($self, $blocking) = @_;
          my $nonblocking = $blocking ? "0" : "1";
          ioctl($self, 0x8004667e, $nonblocking);
      };
   } else { #XXX - do I need this, since caller imports all POSIX constants?
            #XXX - does it hurt?
      require Errno;
      import  Errno qw(EWOULDBLOCK EAGAIN EINPROGRESS ENOBUFS 
              ECONNREFUSED EMFILE EINVAL);
   }
};

sub nonBlockingConnect { #for both Windows *and* unix systems
  #Returns socket handle after connection if timeout undef, or >0
  #Returns socket handle immediately if timeout == 0
  #Always returns $socketError (ok:0, refused:1, timeout:2, 
  #no buffers:3, too many handles:4
  use IO::Socket;
  use IO::Select;
  $socketError = 0;
  my ($addrPort, $timeout) = @_; #$addrPort may be fqdn
  #my ($addr, $port) = split(/:/, $addrPort) or next;
  my ($addr, $port) = split(/:/, $addrPort);
  unless ($addr and $port) {
    print "Bad addrPort sent to nonBlocking connect ($addrPort)\n";
    return undef;
    }

  my $sock = IO::Socket::INET->new(Proto=>'tcp');
  unless ($sock) { 
    if ($! == EMFILE) { #too many handles
      $socketError = 4;
      }
    else {print "unhandled err, sockNew: %d, %s, %s\n", $!, $!, $^E};
    return undef;
    };
  
  defined($sock->blocking(0)) or 
  warn "Couldn't set nonblocking: $^E", return undef;

  $addr = inet_aton($addr) or warn("inet_aton: $!"), return undef;
  my $sockaddr = sockaddr_in($port, $addr) or warn("sockaddr"), return undef;
  if ($debug > 3) {
    print "Sockaddr: $addr:$port -> $sockaddr\n";
    print "and reversed: ";
    my ($p, $a) = sockaddr_in($sockaddr);
    print inet_ntoa($a), ":", $p, "\n";
    };

  my $result = ($sock->connect($sockaddr) or
    ($! == EWOULDBLOCK) or 
    ($! == EINPROGRESS));

  unless ($result) { #potential failure, return undef
    #anything else is an error, probably refused
    #don't warn for common conditions
    if ($! == ENOBUFS) { #ran out of space for sockets
      #XXX - even when this happens, on win2k, the status returned is 
      #XXX - EWOULDBLOCK, so there's no way to know whether we've run 
      #XXX - out or not. There's no choice but to limit the number of 
      #XXX - sockets created to less than 50 or so, or use a canary socket
      #XXX - Might work on some other Windows versions.
      if ($debug > 2) {print "\7Ran out of a system resource for sockets\n"};
      #print "Probably all sockets will fail after this :-(\n";
      $socketError = 3;
      return undef;
      }
    elsif ($! == ECONNREFUSED) {
      if ($debug > 3) {print "Connect to $addrPort (1) refused(?): $!\n"};
      $socketError = 1; #refused
      return undef;
      }
    elsif ($! == EINVAL) {
      if ($debug > 3) {print "Bad socket address $!\n"};
      $socketError = 4; #why does this happen?
      return undef;
      }
    else {  #no result, not blocked, not in progress, not too may filehandles
            #not invalid socket address... ???
      warn sprintf "unknown error: %d, %s, %s\n", $!, $!, $^E;
      print "Error $addrPort; no socket returned from nonBlockingConnect\n";
      return undef;
      };
    };
  #A socket can be returned here which later becomes writable(briefly), 
  #but refused (ECONNREFUSED is set). has_exception also gets set.
  #Caller must handle this for these immediate return sockets.
  if (defined($timeout) and ($timeout == 0) and $result) {return $sock};
  
  #not returned immediately, so we can ensure return of only connected ones
  unless (defined($timeout)) {$timeout = 5};
  my $sel = IO::Select->new($sock) or warn;
  my $start = time;
  my (@sw, @se);
  while (time < ($start+$timeout)) { 
    last if (
      @sw = $sel->can_write(.5) or #can still be refused
      @se = $sel->has_exception(.5)); #this one catches conn refused.
    };
  unless (@sw or @se) { #we dropped through because of timeout
    #warn "timeout";
    $socketError = 2; #timeout
    return undef;
    };
  
  unless ($sock->connected) { 
    #not connected and not timeout, so probably refused
    #get the real (delayed) socket error.
    #100061 on win32.
    if ($debug > 3) {
      $! = $sock->sockopt(SO_ERROR); #to get human-readable error message
      print "Can't connect to $addrPort (2): $!\n"
      };
    $socketError = 1;
    return undef;
    };
    
  return $sock;
  };
  
=item isConnectable(listOfServicesToCheck, timeout);
Check a list of services for connection. Return true/false in scalar 
context, or hash of true/false in list context (closing all sockets 
before return). 

=cut

sub isConnectable {
  #This is pretty gross, but Windows doesn't return errors when 
  #the socket max is exceeded, so there's no alternative to the 
  #'canary' socket below, AFAICS.
  #Advice welcomed!
  #The value varies unpredictably(?), so best to be conservative.
  #Win2k adv server with lotsa RAM is max ~125, so assume all 
  #versions of Windows are less than 150 or so. Unix (and even 
  #Cygwin on Windows) has no limit. Unfortunately, the test below 
  #itself, may decrease available resources.
  #Symptoms of exhausting this resource is that all sockets fail
  #(even those created before and after. So we start with set 
  #size, add a 'canary' socket, reduce the set size if the canary 
  #dies, and remove the canary socket from the returned data 
  #when we have succeeded (canary doesn't timeout, so socket set 
  #size was ok).
  use IO::Socket;
  use IO::Select;
  my ($listToCheckR, $timeout, $flags) = @_;
  #Normally return only 0 (not connectable), or 1 (connectable)
  #but this is changed by $flags.
  #$flags may take the sum of the following values:
  #1: return -1 if socket timed out
  #2: return socket handles (note that if multiple sets are needed, 
  #4: guarantee a return within $timeout
  #   Note that socket handles may already have been closed by the 
  #   remote end, by the time we have waited even one timeout (which 
  #   we may need to do if even one of the attempted connections is 
  #   slow, or if we need to redo the lot because we ran out of 
  #   resources).
  #   Not also that as more handles are accumulated, the usable set size 
  #   will decrease (Windows). 
  my ($returnTimeoutInfo, $returnSocketHandles) = (0,0);
  unless ($flags) {$flags = 0};
  if ($flags&0x0001) {$returnTimeoutInfo++}; #as -1
  if ($flags&0x0002) {$returnSocketHandles++}; #instead of 1
  
  my %connectable = %{$listToCheckR}; #XXX - don't need to make a copy
  
  my $canary;
  my $canaryPort = 1;
  while (!$canary) {
    unless (exists($connectable{"127.0.0.1:$canaryPort"})) {
      $canary = "127.0.0.1:$canaryPort";
      last;
      };
    $canaryPort++;
    };
  my $canaryPresent = 0; #XXX - remove this cruft. Never happens now
  my $canaryTimeout = 1; #wait a max of 1 second for the canary
  #Our canary *might* be in the list to test already. #XXX no, it won't :-)
  #if (exists($listToCheckR->{$canary})) {
  #$canaryPresent++;
  #if($debug>3) {print "\7Canary present!\n"};
  #};
  #XXX - sort by port for debugging ...
  my @list = 
    #sort { 
    #my ($pa, $pb);
    #($pa=$a) =~ s/.*\:(\d+)/$1/; 
    #($pb=$b) =~ s/.*\:(\d+)/$1/; 
    #$pa<=>$pb;
    #} 
    keys(%{$listToCheckR});
  unless ($list[0]) {print "No services to test\n"; return (undef, undef)};
  if ($debug > 3) {print "List to check: @list\n"};
  #and always try for more sockets (the available value seems to 
  #increase after the first try - maybe ram is being allocated).
  #The overall time taken is nrSets*$timeout, so want to make sets 
  #as large as poss. Note that the size determined here is probably 
  #useless to the rest of the application, so we don't change the 
  #global value (we just keep that as an initial value for each entry)
  my $setSize = $maxSocketsAllowed * 2; 
  my $maxSetRetries = 2; #max retry for each set (decreasing set size)
  my $maxTime = $timeout;  #if we're returning socket handles, we can't 
                     #muck around in here for longer than this
                     #Many connections are dropped at 10 seconds even.
                     #Either way, the caller spec'd this, and we honor it.
  #allow for canary too.
  #if (@list < ($setSize+1)) {$setSize = @list+1};
  if ($setSize > @list + 1) {$setSize = @list+1};
  #Note: on a quiet Windows system with plenty of RAM, we can expect 
  #that if this is less than 50 or so, the set size will be ok, the 
  #(connection info|socket handles) will be returned successfully.
  #More than that, and the pathologically broken nature of Windows 
  #implementation of the BSD stack starts to cause problems. Attempts 
  #to retry below, are generally not very successful (and take so 
  #long, that the existing good sockets often get closed by the 
  #remote end, when we are trying to return them to the caller).
  #XXX - can't see any way to hide this from the caller.
  my ($finished, $timedOut);
  my $canaryDied = 0;
  my $listIndex = 0; 
  my $startListIndex = 0;
  my $abort = 0;
  my $totalKept = 0;
  my $startTime = time;
  do {  #sets
    my $setRetry =0;
    do {  #this set, until our canary lives (yes, even for *one* 
          #connection (on Windows)!
      #If he died, redo failed set (with a new reduced set size), 
      #$maxRetries times and try to continue with the rest as 
      #requested
      #if ($debug) {print "loop0\n"};
      my %svc;
      $startListIndex = $canaryDied ? $startListIndex : $listIndex;
      $listIndex = $startListIndex;
      if (($canaryDied) and ($debug > 2)) {
        print "Canary died with socket set size $setSize\n";
        if ($debug > 3) {
          print "Starting at index: $startListIndex\n";
          };
        };

      my $sel = IO::Select->new(); #new one for each new retry

      my $setIndex = 0; 
      my $canarySock = undef;
      my $canaryOk = 0;
      $canaryDied = 0;
      do { #until set filled, or list empty, or can't make more sockets
        #if ($debug>3) {
          #print "Filling set, set index: $setIndex, listIndex: $listIndex\n"};
        my $hostAddrPort;
        if ($setIndex == 0) { #add canary to each set
          $hostAddrPort = $canary;
          #if we don't delete the last canary result, we see it as defined 
          #below, and loop forever (since we don't bump setIndex)
          unless ($canaryPresent) {delete $connectable{$canary}};
          }
        else {
          $hostAddrPort = $list[$listIndex];
          $listIndex++; #read ptr
          };
        #don't test any the caller already has a result for
        undef $socketError; 
        unless (defined($connectable{$hostAddrPort})) {
          if ($debug>3) {print "Connecting to $hostAddrPort\n"};
          #make socket, return socket (or error in global $socketError)
          #XXX - core dumps in Cygwin, if 3600 connections are made
          #XXX - (saw ENOBUFs, sb handled below with $abort???). I think 
          #XXX - IO::Select crashed us(?)
          if (my $sock = nonBlockingConnect($hostAddrPort, 0)) {           
            $sel->add($sock) or warn; 
            $svc{$sock} = $hostAddrPort;
            $canarySock = $sock if ($hostAddrPort eq $canary);
            }
          else {  #unable to make socket - 
                  #probably refused very quickly (socketError == 1)
                  #but maybe out of handles (socketError == 4), 
                  #or buffers (3)
            if ($hostAddrPort eq $canary) {
              if ($socketError == 1) { #canary conn refused quickly
                $canaryOk++; #it worked (didn't timeout)
                if ($debug>3) {print "Canary ok, refused quickly\n"};
                }
              else { 
                #no buffs, or too many handles, etc...
                $canaryDied++;
                if ($debug>3) {
                  print "Canary died,$socketError - too many handles etc.\n"};
                };
              };
            $connectable{$hostAddrPort} = 0;
            if (($debug > 3) and $socketError) { #including canary!
              print "Error connecting to $hostAddrPort: $socketError\n";
              };
            if ($socketError =~ /(4|3)/) {  #out of handles or buffers, 
                                      #handle what we have
                                      #but reduce set size for next time
              if ($debug > 2) {
                print "Too many file handles (have ", 
                  scalar($sel->handles), ") or EINVAL error\n" if 
                    ($socketError == 4);
                print "Too many file buffers (made ", 
                  scalar($sel->handles), ") sockets\n" if ($socketError == 3)
                };
              $maxSocketsAllowed = $sel->handles;
              $abort++;
              };
            };
          $setIndex++; #write ptr
          }
        else {if ($debug>3) {
          print "Skipping $hostAddrPort - already defined\n"};
          };
        } until (($setIndex == $setSize) or 
                ($listIndex == @list) or 
                $abort);

      #end of a set or the list, or we made too many handles
      #handle what we did, reset setIndex
      my $timeStart = time;
      my $timeEnd = $timeStart + $timeout;
      my $nrKept = 0;
      #$canaryDied = 0;
      while ((time < $timeEnd) and $sel->handles and !$canaryDied) {
        if($debug>3) {print "loop1, handles: ", scalar($sel->handles), "\n"};
        #timeout for the canary is different. Don't expect to see him here 
        #if already refused above ($canaryOk set)
        if (!$canaryOk and ((time - $timeStart) > 1)) { #canary has timed out
          $canaryDied++;
          if ($debug>3) {print "Canary timeout\n"};
          };
        for my $sock ($sel->can_write(.2)) { #keep them, or note and close 
          #there will be none in here if the $setSize is too large
          #XXX - If we're keeping the sockets open ($returnSocketHandles)
          #XXX - need to subtract the number we find here from the next 
          #XXX - try at $setSize, otherwise we get a retry 
          #XXX - (and the consequent extra timeout).
          if ($debug > 4) {print "$sock writable\n"};
          #sockets may become writable(!) even though refused (delayed 
          #error is correctly reported as ECONNREFUSED,and connected 
          #returns undef)          
          if (!$sock->connected) {
            #Canary may have become writable, yet not connected. 
            #Check for this (but there's no socket if it was already 
            #refused)
            if (!$canaryOk and ($sock == $canarySock)) {$canaryOk++};
            $connectable{$svc{$sock}} = 0;
            $sel->remove($sock);
            #canary may be last to respond!!!
            $socketError = 1 unless (
              $socketError or 
              ($canarySock and ($sock == $canarySock)));
             #shutdown($sock, 2);
            killSock($sock);
           }
          else {
            $sel->remove($sock);
            $socketError = 0 unless ($canarySock and ($sock == $canarySock));
            if ($returnSocketHandles) {
              $connectable{$svc{$sock}} = $sock;
              $nrKept++;
              }
            else { #avoid keeping refs to socket objects
              $connectable{$svc{$sock}} = 1;
              #shutdown($sock, 2);
              killSock($sock);
              }
            };
          };
        for my $sock ($sel->has_exception(.2)) { 
          if ($debug > 4) {print "$sock has exception\n"};
          if ($sock == $canarySock) {$canaryOk++};
          $connectable{$svc{$sock}} = 0;
          $sel->remove($sock);
          #shutdown($sock, 2);
          killSock($sock);
          #canary may be last to respond!!!
          $socketError = 1 unless (
            $socketError or 
            ($canarySock and ($sock == $canarySock)));
          };
        }; 
      $totalKept += $nrKept; #nr good handles to report to user
      if ($returnSocketHandles) { 
        #reduce our next try to allow for the ones we've kept waiting
        $setSize -= $nrKept;
        #must have at least 2, for even a single socket connection
        # - in Windows.
        if ($setSize < 2) {$setSize = 2};
        };
      #the rest will time out (Windows), or have timed out 
      #(non-Windows) - clean up
      #must find out if the canary died before deciding the values to 
      #return to caller
      #for my $h (@sh) {$canaryDied++ if ($h == $canarySock)};
      if (($debug > 0) and $canaryDied) {
        print "Can\'t reliably create $setSize sockets\n";
        };
      if ($canaryDied) {
        $setSize = int(.7*$setSize);
        #if it's too low, start again (the ok max gets larger with time!).
        if ($setSize < 2) { 
          $setSize = int($maxSocketsAllowed / 2);
          if ($setSize > @list + 1) {$setSize = @list+1};
          #give some a chance to  close
          #unless ($returnSocketHandles) {
            #print "Waiting (60 secs) for sockets to close\n";
            #sleep(60); 
            #};
          };
        };
      my @sh = $sel->handles;
      for my $sock (@sh) {
        unless ($canaryDied) { #leave undef if canary died
          $connectable{$svc{$sock}} = $returnTimeoutInfo ? -1 : 0;
          $socketError = 2 unless $socketError;
          };
        $sel->remove($sock);
        #shutdown($sock, 2);
        killSock($sock);
        };
      #print "WLib socketError: $socketError\n";
      unless ($canaryDied) { #try to keep set size high
        $setSize = int($setSize * 1.2);
        };      
      undef $sel;
      $setRetry++;
      #stop repeating this set if the canary survives, or we've 
      #retried too many times, or we can't create new sockets
      #or we run out of time
      } until (!$canaryDied or $abort or
        (($setRetry > $maxSetRetries) and $totalKept) or
        ((time - $startTime) > $maxTime)); 
    #abort if too many retries and we have something to send caller
    #otherwise leave it to him to notice - 
    #XXX - fix that with a timeout maybe?
    #XXX - should probably count nr of sockets + refused to decide 
    #XXX - whether to return
    $abort++ if (($setRetry > $maxSetRetries) and $totalKept);
    } until ($finished = ($listIndex >= @list) or 
        $abort or 
        ($timedOut = (time - $startTime) > $maxTime));
        #stop filling sets when list empty
  #remove canary unless user wanted it tested(!) ...
  if ($debug) {
    print "Aborted at $listIndex\n" if $abort;
    print "Timed out at $listIndex\n" if $timedOut;
    if ($debug>3) {print "Finished at $listIndex\n" if $finished};
    };
  unless ($canaryPresent) {delete $connectable{$canary}};
  #return whatever we have (many might be dead, not even tried etc.)
  #The caller must handle this. See statProxy for examples.
  #In particular we might be returning with $socketError set to 
  #indicate the error on the *last* socket creation attempt. 
  #socketErrors other than 1 (refused) or 2 (timeout) should normally 
  #not be returned, but may be if we run out of time, or abort because 
  #of too many retries etc.
  return %connectable; 
  };      
  
sub killSock { #close a socket, lose any data, ensure no TIME_WAIT, etc...
               #Need to set so_linger *on*, but set the linger time to 0
               #(if linger is off, the system waits while remaining data 
               #is sent, and then waits 2*MSL in TIME_WAIT)
  my $sock = shift;
  setsockopt($sock, SOL_SOCKET, SO_LINGER, pack("II", 1, 0)) or 
    warn "Can't set SO_LINGER for socket: $!";
  shutdown($sock, 2); #device and queues
  close($sock); #handle
  };

=item $myIPAddress = myIP();
Get this computer's Internet interface IP address. 
Tricky for MS Windows dialups with an internal network interface.

=cut

sub myIP {
  #Get my IP address(es) and decide which to use according to 
  #service connected to (or just connect all on the outside address).
  #Also used to substitute in the customized proxy.pac.
  #we try various ways in various systems to get the result
  
  #Use the first one found in 'ipconfig', or localhost for win9*.
  #Try the code from master.pl here.

  #XXX better code (no qx for win98 KSA arabic boxen) is:
  #use Sys::Hostname;
  #$Name2=hostname();
  #print "Hostname: $Name2: \n";
  #for $addr (@{gethost(lc($Name1))->addr_list}) {
    #print join(".", unpack(C4, $addr )),"\n";
    #};
  #This gives a list of addresses; private ones ignored.
  #If there's none left, need to find the default gateway
  #If there's more than one left use nonBlockingConnect
  my $os = shift;
  unless ($os) {$os = getOS()};
  my $myIpStr='127.0.0.1';
  if ($os =~ /^1$/) { #XXX - this one might have cygwin too
    my $rt = qx/route print/;
    $myIpStr = ($rt =~ 
     m|0\.0\.0\.0.*?0\.0\.0\.0.*?\d+\.\d+\.\d+\.\d+.*?(\d+\.\d+\.\d+\.\d+)|sg)[-1];
    }
  elsif (($os =~ /^2$/) or ($^O =~ /cygwin/)) {
    $myIpStr = (qx/ipconfig/ =~ m|IP Address.*?(\d+\.\d+\.\d+\.\d+)|sg)[-1]
    }
  elsif ($os=~ /^10$/) {
    my $hn = qx/hostname/; chomp($hn);
    my $nsl = qx/nslookup $hn 2>nul/;
    ($myIpStr) = $nsl =~ m|Name:.*?Address:\s*(\d+\.\d+\.\d+\.\d+)|sg;
    unless ($myIpStr) { #need to use ifconfig, if we can find it
      my $ifStr = `ifconfig`; #shows all interfaces in linux, but not netBSD
      my @ipStrs = $ifStr =~ /inet\s(?:addr:)?(\d+\.\d+\.\d+\.\d+)/ig;
      for my $ip (@ipStrs) { #find a valid internet address (allow private ones)
        next if ($ip =~ /^127\./);
        $myIpStr = $ip;
        #XXX - need to handle the /12 rfc1918 address better below
        last unless ($myIpStr =~ /^(192\.168|172\.)\./) ; #maybe a better one
        };
      };
    };
  if ($myIpStr) {
    if ($debug) {print "I\'m assuming this IP address is $myIpStr\n"};
    }
  else {
    print "You seem to have no IP address - are you connected?\n";
    print "I will assume it is 127.0.0.1\n";
    $myIpStr='127.0.0.1';
    };
  return $myIpStr;
  };

=item $nameServers = getNameServers();
Get the nearest name servers to the host running this proxy tool app

=cut

sub getNameServers { #collect name servers from the o/s
  #note: can get a name server by a whois lookup if necessary.
  #win9* doesn't have nslookup, so need to know o/s with better 
  #granularity than $^O. Update: they both have ipconfig /all, 
  #so getOs granularity is ok.
  #for unix, we can get the name servers from /etc/resolv.conf 
  #(easiest to make  Net::DNS::resolver and look at what it 
  #finds in $res->nameservers; that doesn't work in Windows)
  #XXX - getOS returns different values!
  #XXX - POSIX::uname is probably better.
  #use File::Spec;
  my %nameServers = ();
  if ($network->{provider}->{nameServer}) { #keep anything we already have ...
    @nameServers{split(/,\s*/, $network->{provider}->{nameServer})} = ();
    };
  my %ifaceAddresses = ();
  if ($network->{global}->{interfaceAddress}) { #keep anything we already have ...
    @ifaceAddresses{split(/,\s*/, $network->{global}->{interfaceAddress})} = ();
    };
  
  my $osVer = getOS();
  if ($debug > 3) {print "O/S: $osVer\n"};
  
  if (($osVer > 5) and ($^O !~ /cygwin/)){ #Unix, linux etc.
    #XXX - could use /etc/resolv.conf, $HOME/.resolv.conf
    #XXX - ./.resolv.conf, dig
    my $nullDev = $^O =~ /Win32/ ? 'nul' : '/dev/null';
    #my $nullDev = File::Spec->devnull(); #maybe need this for Macs?
    my @nameServer=`nslookup a 2>$nullDev`;
    if ($nameServer[1] =~ /Address:\s*(\d+\.\d+\.\d+\.\d+)/) {
      #$network->{provider}->{nameServer} = $1;
      $nameServers{$1} = ();
      #print $1;
      }
    }
  else {  #use ipconfig /all for winNT4, NT5.0(w2k), NT5.1(wxp), 
          #winME (95/98?), Cygwin
    my $IPConfig = `ipconfig /all`;
    if ($IPConfig) { #some win98 systems return null ???
      my @nameServers = ($IPConfig =~ 
        /[\r\n]+[\s\t]*DNS\sServers[^:]*:
        [\s\t]*(\d+\.\d+\.\d+\.\d+)[^\r\n\d]*
        [\r\n]+
        [\s\t]*(\d+\.\d+\.\d+\.\d+)?[^\r\n\d]*/gx);
        #remove dupes (supressing undef errors)
      if (@nameServers) {$^W=0; @nameServers{@nameServers} = (); $^W=1}; 
      delete $nameServers{''}; #if an array entry existed, but was undefined
      if (%nameServers) {
        if ($debug>3) {
          print "Name servers from ipconfig, nslookup etc.: ", 
            join(', ', keys(%nameServers)), "\n";
          };
        }
      else {
        $network->{provider}->{nameServer} = undef;
        print "Can't find your name server(s) using nslookup, ipconfig /all etc.\n";
        };
      };
    #try the registry, but not if it's Cygwin (don't want Cygwin 
    #using Win32 modules)
    unless ($osVer > 5) { #'specially not for Cygwin
      #all except ME first ...
      require Win32::TieRegistry;
      my $tcpKey;
      if ($tcpKey = $Win32::TieRegistry::Registry->
        {"HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\".
          "Tcpip\\Parameters"}) {
        #scan for adapter info
        #print $tcpKey, "\n";
        for my $adapterType (keys(%{$tcpKey})) { 
          #Win2k|WinXP cases. Note that a clean NT4 w/s has none of these 
          #entries. They are all under the nic device entries
          #XXX - fix that.
          next unless ($adapterType =~/(DNSRegisteredAdapters|Interfaces)/i); 
          for my $adapter (keys(%{$tcpKey->{$adapterType}})) { #each interface
            next unless ($tcpKey->{$adapterType}->{$adapter});
            for my $adapterInfo (keys(%{$tcpKey->{$adapterType}->{$adapter}})) {
              next unless $adapterInfo;
              my $valData = $tcpKey->{$adapterType}->{$adapter}->{$adapterInfo};
              if ($adapterInfo =~ /(DNSServerAddresses|NameServer)/i) { 
                my @regNameServers;
                if ($valData =~ /\d+\.\d+\.\d+\.\d+/) { #it's a string
                  @regNameServers = split(/[\s,]\s*/, $valData);
                  }
                else { #it's probably an octet list
                  @regNameServers = 
                    (join('.', unpack("C*", $valData)) =~ /\d+\.\d+\.\d+\.\d+/g);
                  };
                @nameServers{@regNameServers} = (); #merge into our list ...
                if ($debug>3) {
                  print "Reg dns addresses: ", join(', ', 
                    @regNameServers), "\n"
                  };
                }
              #keep the fixed interface addresses while we are here ...
              elsif ($adapterInfo =~ /RegisteredAddresses/i) { 
                my $regCount;
                if (defined($tcpKey->{$adapterType}->{$adapter}->
                  {'RegisteredAddressCount'})) {
                  $regCount = eval($tcpKey->{$adapterType}->{$adapter}->
                    {'RegisteredAddressCount'}); #val is like "0x00000001"
                  }
                else {$regCount = 0};
                if ($regCount) {$valData = substr($valData, 0, 4*$regCount)};
                my @regIfaceAddrs = 
                  (join('.', unpack("C*", $valData)) =~ /\d+\.\d+\.\d+\.\d+/g);
                if ($debug > 3) {
                  print "Fixed interface addresses: ", 
                    join(', ', @regIfaceAddrs), "\n";
                  };
                @ifaceAddresses{@regIfaceAddrs} = (); #merge into our list ...
                }
              elsif ($adapterInfo =~ /IPAddress/) { 
                if (defined(
                    $tcpKey->{$adapterType}->{$adapter}->{IPAddress})) { 
                  my @regIPs =  
                    $tcpKey->{$adapterType}->{$adapter}->{IPAddress} =~ 
                      /(\d+\.\d+\.\d+\.\d+)/g;
                  # even allow 0.0.0.0 - handled by caller
                  if (@regIPs) {
                    @ifaceAddresses{@regIPs} = ();
                    if ($debug > 3) {
                      print "IP addresses from reg: ", 
                        join(', ', @regIPs), "\n";
                      }
                    };
                  }
                };                  
              };
            };
          };
        } #end if $tcpKey
      #next - try the WinME style registry structure <groan>
      if ($tcpKey = $Win32::TieRegistry::Registry->
        {"HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\".
          "VxD\\MSTCP"}) {
        for my $adapterInfo (keys(%{$tcpKey})) { 
          #WinME case
          next unless ($adapterInfo =~ /NameServer/i); 
          my $valData = $tcpKey->{$adapterInfo};
          #my @regNameServers = 
            #(join('.', unpack("C*", $valData)) =~ /\d+\.\d+\.\d+\.\d+/g);
          #if ($debug > 2) {
            #print "Reg dns addresses: ", join(', ', 
              #@regNameServers), "\n";
            #};
          #@nameServers{@regNameServers} = (); #merge into our list ...

          my @regNameServers;
          if ($valData =~ /\d+\.\d+\.\d+\.\d+/) { #it's a string
            @regNameServers = split(/[\s,]\s*/, $valData);
            }
          else { #it's probably an octet list
            @regNameServers = 
              (join('.', unpack("C*", $valData)) =~ /\d+\.\d+\.\d+\.\d+/g);
            };
          @nameServers{@regNameServers} = (); #merge into our list ...
          if ($debug>3) {
            print "Reg dns addresses: ", join(', ', 
              @regNameServers), "\n"
            };
          };
        };
      }; #end unless Cygwin
    }; #end if unix, cygwin {} else windows
    
  $network->{provider}->{nameServer} = join(', ', keys(%nameServers));
  if ($debug>2) {
    print "DNS addresses from o/s: ", $network->{provider}->{nameServer}, "\n"
    };  
  if ($osVer <5) { #we did this for Windows only
    $network->{global}->{interfaceAddress} = join(', ', keys(%ifaceAddresses));
    if ($debug>3) {
      print "Fixed interface addresses: ", 
        $network->{global}->{interfaceAddress}, "\n"
      };
    };
  return wantarray ? keys(%nameServers) : $network->{provider}->{nameServer};
  };

sub getNameServers1 { #use our IP address to find the name server(s) 
                      #providing the reverse lookup, if any.
  my $fqdn = gethostbyaddress(myIP(getOS())); #doesn't work often enough!
  print "fqdn: $fqdn\n";
  my $domain = (split(/\./, $fqdn, 2))[1];
  my $res = Net::DNS::Resolver->new();
  my $query = $res->query($domain, "NS");
  if ($query) {
      foreach my $rr ($query->answer) {
          next unless $rr->type eq "NS";
          print $rr->nsdname, "\n";
        }
    }
  else {
      print "query failed: ", $res->errorstring, "\n";
    };
  };

sub getNameServers2 {#use our IP address to find the name server(s) 
                    #we should be using.
                    #If all else fails, use the ones providing the 
                    #reverse lookup, if any.
                    #To do that, we need a name server and there 
                    #may not be a default one available. In that case 
                    #we use a root NS.
  #XXX - if this is bullet proof, use it to replace the corresponding 
  #XXX - code in master.pl
  my $timeout = 20;
  my ($dnsFqdn, $dnsAddr, $level);
  #Net::DNS gets resolver name servers from /etc/resolv.conf
  #On MSWin systems, this is not usually present, and the name server 
  #defaults to 127.0.0.1 (does *not* look in the registry).
  my $res = Net::DNS::Resolver->new();
  $res->recurse(0);
  my $os = getOS();
  #print $res->nameservers, "\n";
  if (!defined($res->nameservers()) or 
      (defined($res->nameservers()) and 
      #MS Windows uses 127.0.0.1 even when it has no DNS!
      ($os < 10) and 
      (($res->nameservers())[0] =~ /^127/))) {
    $res->nameservers('128.8.10.90'); #a root server
    };
  if ($debug>2) {print "default NS: ", $res->nameservers(), "\n"};
  my $myIpAddr = myIP($os);
  my @myIpAddr = split(/\./, $myIpAddr);
  $level = -1;
  while ($level < 3) {
    my $skt = $res->bgsend($myIpAddr); #will do PTR 
    #my $skt = $res->bgsend('217.in-addr.arpa.'); #will do PTR 
    my $sel = new IO::Select($skt);
    if ($sel->can_read($timeout)) {
      #we can only have $skt here
      my $packet = $res->bgread($skt);
      if (($debug>4) and $packet) {$packet->print};
      if ($packet) { 
        #not interested in the answers, just the authority 
        #(and the IP address of those in the additional section).
        #note - only using the first authority at each level. 
        #XXX - would be more likely to find an answer if i used them all
        #XXX - and backtracked on timeout/error
        my $dnsAuth;
        foreach my $rr0 ($packet->authority) {
          my $type0=$rr0->type;
          if (
            (($type0 eq "NS") and ($dnsFqdn = $rr0->nsdname)) or 
            (($type0 eq "SOA") and ($dnsFqdn = $rr0->mname))
            ) {
            $dnsAuth = $rr0->name;
            #print "dns auth: $dnsAuth\n";
            $level = findLevel($dnsAuth); #how close we are to home
            next unless (defined($level)); #not even the right type!
            if ($debug>3) {
              print "Name server: $dnsFqdn, authority name: $dnsAuth, level: $level\n";
              };
            #XXX - we don't need to do this normally - change server to 
            #XXX - the $dnsFqdn we already have, will resolve using 
            #XXX - the existing one (if there is an existing one, of course).
            undef $dnsAddr;
            #check through these to resolve the $dnsFqdn to an IP address
            foreach my $rr1 ($packet->additional) { 
              next if ($rr1->type ne "A");
              if ($rr1->name =~ /^$dnsFqdn$/i) {
                $dnsAddr = $rr1->address;
                last;
                };
              };
            #next unless $dnsAddr;
            };
          return $dnsAddr if ($level == 3); #finished!
          #return $dnsAddr if ($rr0->name =~ #finished!
            #/"$myIpAddr[2].$myIpAddr[1].$myIpAddr[0].in-addr.arpa"/i);
          if ($dnsFqdn and $dnsAddr) {
            $res->nameservers($dnsAddr);
            if ($debug>3) {print "Found $dnsAddr for $dnsAuth, level $level\n"};
            last;
            }
          else { #the IP addr of the dns server was not in 
                #the additional records, so we must get it
            #print "nameservers: ", $res->nameservers, "\n";
            $dnsAddr = resolveIP($dnsFqdn, 10, $res->nameservers);
            if ($dnsAddr) { #got it
              $res->nameservers($dnsAddr);
              last;
              }
            else { #hmm ... try the coded default NS list
              $dnsAddr = resolveIP($dnsFqdn, 10); 
              unless ($dnsAddr) {next};
              }; #try the next auth record
            };
          };
        #print "dnsFqdn: $dnsFqdn, dnsAddr: $dnsAddr\n";
        return undef unless $dnsAddr;
        }
      else {print "no response\n"; return undef}; # no packet response at all
      }
    else {print "timeout\n"; return undef}; #timeout
    };
  print "unable to find name servers\n";
  return undef;
  };

sub findLevel { #level 0 is a root nameserver, level 3 is a NS 
                #authoritative for c.b.a.in-addr.arpa
  my $dnsFqdn = shift;
  my $level;
  #remove the terminating '.'
  $dnsFqdn =~ s/\.$//;
  my @addr = reverse(split(/\./, $dnsFqdn)); 
  return undef if ($addr[0] !~ /^arpa$/i);
  return undef if ($addr[1] !~ /^in-addr$/i);
  for ($level=0; $level<4; $level++) {
    return $level unless (
      defined($addr[$level+2]) and 
      #don't need to match the myIp address digits, I think
      ($addr[$level+2] =~ /^\d+$/));
    };
  };  
  
=item parseNetworkInfo();
Get as much info about the network as we can

=cut

sub parseNetworkInfo {
  my (%routes, @routes);
  my ($myAddress, $cmd, $IPaddr);
  #parse routing table on Win32, hostname' on *nix
  #get from WLib as well, to compare with the routing table based 
  #guess from further down later 
  my $WLibIPAddressGuess = myIP(getOS()); 
  #my $os = (($^O =~ /(MSWin32|cygwin)/i) ?1 :0);
  for ($^O) { 
      /(MSWin32|cygwin)/ and do {
          my $cmd='route print';
          my @routes=`$cmd`; my $table=0;
          foreach my $route (@routes) {
              chomp($route);
              #print "route: $route\n";
              my ($dest, $mask, $gate, $iface, $metric);
              (undef, $dest, $mask, $gate, $iface, $metric) = 
              #($os ?(undef, $dest, $mask, $gate, $iface, $metric) 
              #:($dest, $gate, $mask, undef, $metric, undef, $iface)) = 
              split(/\s+/, $route);
              #print "dest: $dest\n";
              next unless ($dest);
              if ($route =~ /(Destination|Address)/i) {$table=1};
              next unless ($table);
              if ($table and (($route=~/Default/i) or ($route =~ /====/))) {last};
              next unless ($dest =~ /\d+\.\d+\.\d+\.\d+/);
              $routes{$dest}->{mask} = $mask;
              $routes{$dest}->{gate} = $gate;
              $routes{$dest}->{iface} = $iface;
              $routes{$dest}->{metric} = $metric;
              };
          unless ($table) {warn "can't parse your routing table\n"};
          if ($table) {
            my $minMetric=100; 
            #$defaultGateList='';
            $myAddress='';
            #need to make some guesses about what subnets are acccessible from here - 
            #There is not enough info from the routing table to do this completely.
            #Is there even enough to make a good guess? No.
            #In the general case, only the ISP/corp can know this.
            #OTOH, we can't go testing every proxy we know of at this time, so we 
            #must make a guess, even if we say there are no subnets accessible.
            #Compromise, until I get smarter ... all 
            #(routing table dest addresses)/(their routing table mask) subnets are 
            #accessible. 
            #Nothing else, not even private address nets.
            #
            #Note: if there are addresses we know we can get to, and we want LP to 
            #see them, we can add a routing table entry (which probably should be 
            #there anyway).
            for my $dest (keys(%routes)) {
              $network->{session}->{destinationMasks}->{$dest} = $routes{$dest}->{mask};            
                if ($dest =~ /0\.0\.0\.0/) {
                    if ($routes{$dest}->{metric} < $minMetric) {
                        $network->{session}->{defaultGate} = $routes{$dest}->{gate};
                        $network->{session}->{defaultIface} = $routes{$dest}->{iface};
                        $minMetric = $routes{$dest}->{metric};
                        next;
                        };
                    };
                next if ($dest =~ /\d+\.\d+\.\d+\.(0|255)/);
                next if ($dest =~ /^127\..*/);
                if ($routes{$dest}->{gate} =~ /^127\..*/) {
                    $myAddress .= "$dest, ";
                    };
                };
            if (!defined($network->{session}->{defaultIface})) {
                $network->{session}->{defaultIface} = ''};
            if (!defined($network->{session}->{defaultGate})) {
                $network->{session}->{defaultGate} = ''};
            ##
            #see tcp-ip-info.pl and getIp.pl (both from Dave Roth) 
            #and remember ipconfig/winipcfg
            ##
            }
          #XXX - there's some registry nic examination code in 
          #XXX - getNameServers (for now). Move soon.
          #XXX - It will init $network->{global}->interfaceAddress}
          getNameServers(); 
          };

      /linux/ and do {
          $cmd = 'route -n';
          my $host = `hostname`; chomp($host);
          my $addr = `nslookup $host 2>nul`; chomp($addr);
          $addr =~ /^.*Address.*?(\d+\.\d+\.\d+\.\d+).*$/s;
          $myAddress = $1;
          $network->{session}->{defaultIface} = $myAddress;
          };

      /netbsd/ and do {
          #print "os: netbsd\n";
          $cmd = 'netstat -r';
          my $host = `hostname`; chomp($host);
          #print "host: $host\n";
          my $addr = `nslookup $host 2>nul`; 
          #print "lookup: $addr\n";
          $addr =~ /.*$host.*Address.*?(\d+\.\d+\.\d+\.\d+).*/s;
          $myAddress = $1;
          #print "myAdd: $myAddress\n";
          $network->{session}->{defaultIface} = $myAddress;
          };
      };

  for ($network->{global}->{interfaceAddress}) {$_= '' unless $_};
  for my $dest (split(/,\s* /, $myAddress)) {
      unless ($dest =~ /$network->{session}->{defaultIface}/) {
          $network->{global}->{interfaceAddress} .= "$dest, ";
          };
      };
  #print "Your computer's address(es): \n";
  #print "\thardware interface(s): $network{global}->{interfaceAddress}\n";
  #print "\tsoftware interfaces: $network{session}->{defaultIface}\n";
  #print "Your default gateway interface address: $network{session}->{defaultIface}\n";
  #print "Your default gateway: $network{session}->{defaultGate}\n";

  #pick the first dialup interface address 
  #(unlikely to be more than one)
  #if not defined, choose the first fixed interface address
  $IPaddr = ((split(/,\s*/, $network->{session}->{defaultIface}))[0] or 
      (split(/,\s*/, $network->{global}->{interfaceAddress}))[0]);
  if (!$IPaddr) {
      print "\n\7Can\'t determine your IP address\n"; 
      return undef
      };

  #use Win32;
  #my $MSComputerName = Win32::NodeName();
  #print "Your computer name is $MSComputerName\n"; #MS computer name

  ##
  #see tcp-ip-info.pl and getIp.pl (both from Dave Roth) 
  #and remember ipconfig/winipcfg
  ##

  #to set up the global info, we basically assume we're on a dialup
  #for now
  $network->{global}->{interfaceAddress} = '';
  for my $dest (split(/ /, $myAddress)) {
      unless ($dest=~/$network->{session}->{defaultIface}/) {
          $network->{global}->{interfaceAddress} .= "$dest, ";
          };
      };

  getNameServers(); #adds all of them to %network
  if ($debug>2) {
    print "Your DNS name servers: $network->{provider}->{nameServer}\n"
    };
  return $network;
  };

=item $os = getOS();
Get operating system

=cut

sub getOS { #allocate a number for the OS
  #XXX - POSIX::uname does this portably.
  my $os;
  if (($^O =~ /MSWin32/) and eval(require Win32)) {
    #following is 0 for win32s, 1 for win9*, 2 for nt4, 5, win2k, winXP
    #[1] (major version number) is 4 for NT4, 5 for NT5 (win2k, XP)
    #[2] (minor version number) is 0 for win2k, and 1 for winXP if I 
    #ever need that granularity
    $os = Win32::GetOSVersion()} 
  else {$os = 10}; #call (linu|uni)x *and cygwin*, number 10
  return $os;
  };

=item unzip($fileIn, $fileOut);
Decompress zipped file

=cut

sub unzip { #unzip a file
  my ($fileIn, $fileOut) = @_;
  open IN, $fileIn or croak "open error on $fileIn: $!";
  binmode IN;
  open OUT, ">$fileOut" or croak "create error on $fileOut: $!";
  #binmode OUT;
  #following stuff was needed for Zip archives ... didn't work.
  #my $wbits = MAX_WBITS; $wbits = -$wbits;
  #print "wbits: $wbits\n";
  #my $x = deflateInit( '-WindowBits' => -$wbits )
  my $x = deflateInit()
     or croak "Cannot create a deflation stream for unzipping $fileIn\n";
  my ($output, $status);
  while (<IN>)  {
    ($output, $status) = $x->deflate($_);
    $status == Z_OK
      or croak "deflation failed\n";
    print OUT $output;
    };
  ($output, $status) = $x->flush();
  close IN;
  close OUT;
  return $status == Z_OK;
  };

=item unarchive($fileIn, $fileOut);
Unarchive and decompress zip archive

=cut

sub unarchive { #unarchive and decompress a file from a zip archive
  my ($archFile, $outFile) = @_;
  my $zip = Archive::Zip->new($archFile);
  #extract outFile from archive archFile, overwriting any outFile already there.
  unless ($zip->extractMemberWithoutPaths($outFile, $outFile) == Z_OK) { 
    print "error unzipping $archFile.$outFile: $!\n";
    return undef;
    };
  return 1;
  };

=item firewall($addr, $firewallsHashRef)
Return the firewall (name) the given host is in

=cut

sub firewall {  #find firewall that this IP addr is in (used 
                #in mergeHosts to avoid testing acl'd or firewalled 
                #hosts) and merging the results into hosts.xml
  my ($addr, $firewalls) = @_;
  my $str;
  return undef unless (
    $addr and 
    $firewalls
    );
  my (@fw, @mask);
  for my $fw (keys(%{$firewalls})) {
    next if (
      !$fw or 
      ($fw =~ /firewall-none/) or #cant allow this one!
      (ref($firewalls->{$fw}) !~ /HASH/) or 
      !defined($str = $firewalls->{$fw}->{subnetsInside})
      );
    for my $subnet (split(/, /, $str)) {
      my $mask = (split(/\//, $subnet))[1];
      if (isInSubnet($addr, $subnet)) {
        push(@fw, $fw);
        push(@mask, "$mask.$fw");
        };
      };
    };
  if (@fw) {
    if (@fw>1) {#order by specificity
      #print join(' ', @mask);
      @mask = sort {(split(/\./, $b))[0] <=> (split(/\./, $a))[0]} @mask;
      @fw = map {(split(/\./, $_, 2))[1]} @mask;
      };
    return wantarray ? @fw : $fw[0]; #most specific one found
    }
  else {
    return undef; #not in any *known* firewall.
    };
  };

sub isInSubnets { #is $addr in the string of subnets specified 
  my ($addr, $subnets) = @_;
  for my $subnet (split(/,\s*/, $subnets)) {
    return 1 if (isInSubnet($addr, $subnet));
    };
  return 0;
  };

=item isInSubnet($addr, $subnetMask);
Returns true if $addr is in the subnet $subnetMask (specified like 
a.b.c.d/e (example: 192.168.0.0/24)). Will resolve a.b.c.d to IP 
address. 

=cut

sub isInSubnet {  #Is IP addr in subnet?
                  #do DNS in here - most hosts don't get here 
                  #and those that do never get here more than once.
                  #XXX check this ...
  #XXX - this needs to be upgraded to be able to determine whether 
  #XXX - $addr/$mask is inside $biggerSubnetAddr/$biggerMask. 
  #XXX - $mask may be 32, of course.
  my ($addr, $subnet) = @_;
  return 1 if ($subnet =~ /^all$/); 
  return 0 if ($subnet =~ /^none$/);
  my $dottedAddr = $addr;
  my $sAddr;
  #subnetBase need not have zeros in the bits for the inside host addresses
  my ($subnetBase, $mask) = split(/\//, $subnet);
  unless (($subnetBase=~/\d+\.\d+\.\d+\.\d+/) and 
    ($mask=~/\d+/)) { 
    if ($debug>2) {
      print "Bad subnet: ", defined($subnetBase) ? $subnetBase : '?';
      print "\/", defined($mask) ? $mask : '?', ", addr: $addr, "; 
      print "subnet: ", defined($subnet) ? $subnet : '?', "\n";
      };
    return undef;
    };
  my $bitMask = 0;
  for (my $bit=31;$bit>=0;$bit--) {
    $bitMask = ($bitMask<<1) + ($bit > (31-$mask));   
    };
  #printf "mask: %x\n", $bitMask;
  unless ($addr =~ /^\d+\.\d+\.\d+\.\d+$/) { #need dns
    unless ($dnsInstalled) {
      print "Warning: no DNS installed; so I can't help you with\n";
      print "the use of hosts without IP addresses in the hosts\n";
      print "or config-*.xml files\n";
      print "Please install it, or use only IP addresses in config files\n";
      return undef;
      };
    #resolve IP address $dnsTimout timeout
    $addr = resolveIP($addr, $dnsTimeout) or do {
      print "DNS name resolution failed\n";
      return undef
      };
    if ($debug>4) {
      print "Resolved $dottedAddr to $addr\n";
      croak "Bad address: $addr" unless ($addr =~ /\d+\.\d+\.\d+\.\d+/);
      };
    };
    
  my @addr = split(/\./, $addr); 
  unless (defined($addr[3])) {
    if ($debug>1) {print "bad address: $dottedAddr, $addr\n"};
    return undef; #fail 'isInside'
    };
  $addr = $addr[0];
  for my $i (1..3) {
    $addr = ($addr<<8) + $addr[$i];
    };
  #printf "addr: %x\n", $addr;
  my @sAddr = split(/\./, $subnetBase); 
  croak unless defined($sAddr[3]); #if this is wrong, it needs fixing.
  $sAddr = $sAddr[0];
  for my $i (1..3) {
    $sAddr = ($sAddr<<8) + $sAddr[$i];
    };
  #printf "sAddr: %x\n", $sAddr;
  return (($addr & $bitMask) == ($sAddr & $bitMask));
  };

=item isAccessible();
Check if a host is accessible and allows access

=cut

sub isAccessible { #check if host addrPort is accessible from LPHost 
                  #(where we are). Take into account outgoing firewall 
                  #blocks from $origAddrPort, $destAddrPort incoming 
                  #firewall blocks, and dest service ACL lists.
  my ($origAddrPort, $destAddrPort, $hostR, $firewalls) = @_;
  my ($origAddr, $origPort) = split(/:/, $origAddrPort);
  my ($destAddr, $destPort) = split(/:/, $destAddrPort);
  my $origFw = (firewall($origAddr, $firewalls) or 'unknown');
  my $destFw = (firewall($destAddr, $firewalls) or 'unknown');
  if ($origFw eq $destFw) { #inside same firewall
    return destSvcPermits($origAddr, $origPort, $destAddr, $destPort, $hostR);
    }
  else {
    return 
    (origFwPermits($origAddr, $origPort, $destAddr, $destPort, 
      $firewalls->{$origFw}->{outgoing}) and 
    destFwPermits($origAddr, $origPort, $destAddr, 
      $destPort, $firewalls->{$origFw}->{outgoing}) and 
    destSvcPermits($origAddr, $origPort, $destAddr, $destPort, $hostR));
    };
  };
  
sub origFwPermits { #firewall at originating end of this TCP conection does 
                    #not block it
  my ($origAddr, $origPort, $destAddr, $destPort, $fwR) = @_;
  if  (defined($fwR->{blockedTcpPorts}) and 
      ($fwR->{blockedTcpPorts} =~ /(^|,\s*)$destPort(,\s*|$)/)) {
    return 0
    }
  elsif (defined($fwR->{blockedTcpAddresses}) and 
        ($fwR->{blockedTcpAddresses} =~ 
          /(^|,\s*)$destAddr(,\s*|$)/)) {
    return 0
    }
  else {return 1};
  };
  
sub destFwPermits { #firewall at destination end of this TCP connection does 
                    #not block it
  my ($origAddr, $origPort, $destAddr, $destPort, $firewalls) = @_;
  #firewall 'onlyAllowsTcpAccessFrom' if present.
  };
  
sub destSvcPermits { #service (or host-based firewall) at destination 
                    #end of this TCP connection does 
                    #not block it
  my ($origAddr, $origPort, $destAddr, $destPort, $hosts) = @_;
  #host 'onlyAllowsTcpAccessFrom', if present.
  };
  
=item isOnLine();
Check if we are on-line (parallel connections)

=cut

sub isOnLine { #check if we are on-line, update global $onLine var 
  use IO::Socket;
  use IO::Select;
  my ($listToCheckR, $actionR) = @_; 
  my $timeout = 10; #max time to wait for a connect  
  my %connectable = %{$listToCheckR}; #don't put results in original list
                                      #or we will never test them again!
  return 0 unless %connectable; #probably no layer 0 hosts
  %connectable = isConnectable(\%connectable, $timeout, 0);
  return if $exit; #let's not stay here if the user is trying to get out
  #see if one connected
  $onLine = 0;
  for (keys(%connectable)) {
    $onLine++ if (defined($connectable{$_}) and ($connectable{$_} > 0));
    };
  return $onLine; 
  };

=item refreshIp($trustedAddressHashRef);
Check if we have changed our IP address (dynamic dialups etc.)

=cut

sub refreshIp { 
  #in case a dialup, dynamic IP address user has gone off-line 
  #and is returning with a new address. We do this at both on->off 
  #and off->on transitions, so the user can see he has been 
  #disconnected. 
  my $trustedAddressesR = shift;
  #if ($onLine != $ol) {$myIpStr = myIP($os)};
  #nup - we need to do this *all* the time.
  my $newIpStr;
  if ( $myIpStr and 
      ($newIpStr = myIP($os)) ne $myIpStr) {
    if ($debug>2) {print "new IP: $newIpStr, old IP was: $myIpStr\n"};
    #if our old address was trusted, we just replace it in the 
    #trusted addr list; otherwise clear the list and wait to trust 
    #the  next connection (!)
    #XXX - an embedded URI command with username and password 
    #XXX - is probably the 'secure' way to handle this.
    unless ($$trustedAddressesR =~ s/\Q$myIpStr\E/$newIpStr/) {
      print "Changed IP address to $myIpStr; clearing trusted addresses\n";
      print "Note: first connection will take control now!\n";
      undef $$trustedAddressesR;
      };
    $myIpStr = $newIpStr;
    };
  };

=item stringify(string);
make data printable

=cut

sub stringify { #make data printable
  local $_ = shift; 
  s/([\'\"\\])/\\$1/g; #"' for my editor's syntax highlighting
  #enclose the printed string in quotes - can execute as perl stmt
  s/(.*)/"$1"/s; 
    #unless /^(?: -?[\d.]+ | \*[\w:]* )$/x;
  s/([\200-\377])/sprintf("M-%c",ord($1)&0177)/eg;
  s/([\0-\37\177])/sprintf("\\c%c",ord($1)^64)/eg;
  return $_;
  };

=item getXML(file name);
Load an XML file as a hash, return pointer

=cut

sub getXML {
  my ($file) = @_;
  #return undef unless (
    #XXX - need to export these from WLib if we want to use them here
    #$main::xmlParserInstalled and 
    #$main::xmlDumperInstalled
    #);
  my ($base) = $file =~ /^(.+).xml$/;
  my $oldFile = $base.'.old.xml';
  if (-e $base.'.zip') {  #always use the xml file in a zipped 
                          # (latest) version - if one is there
    #XXX - unless it's the same file inside!
    print "Renaming $file to $oldFile, unarchiving $base.zip; \n";
    if (-e $file) {rename($file, $oldFile)};
    if (unarchive($base.'.zip', $file)) {
      #now rename this archive, so we don't do this every time
      #we clobber any existing old zip
      rename($base.'.zip', $base.'.old.zip')
      }
    else {
      print "Error unarchiving $base.zip;using old $file\n";
      };
    };
  croak "No data file: $file" unless (-e "$file");
  open(XMLh, "<$file") or croak "Can't open the data file: $file";
  if ($debug>2) {print "Importing $file "};
  my @xml = <XMLh>; 
  my $xmlStr = join('', @xml);
  close(XMLh);
  
  my $parser = XML::Parser->new(Style => 'Tree');
  my $tree = $parser->parse($xmlStr);
  my $xs = new XML::Dumper;
  my $data = $xs->xml2pl($tree);
  if ($debug>2) {
    if ($data) {
      if (defined($data->{'version'})) {
        print "version ", $data->{'version'} =~ /.*?Revision:\s+(\d+\.\d+)/, 
          "\n";}
      else {print "version unknown\n";}
      }
    };
  unless ($data) {print "failed to load $file\n"};
  return $data;
  };

=item putXML(hash reference, file name);
Save hash as XML file

=cut

sub putXML { #write hash to xml file
  my ($hashR, $file) = @_;
  #local $/ = "\r\n"; #always want this(?), even on Unix systems, 
                    #at least, to standardize. I use this under Cygwin, 
                    #so it writes a Unix format hosts.xml.
                    #By default perl writes the o/s format text file.
  my $xs = new XML::Dumper;
  my $xml =  $xs->pl2xml($hashR);

  open(OUT, ">$file") or croak "can't create $file: $!";
  print OUT $xml;
  close OUT;
  };
  
=item getURL(url, timeout);
Get web page 

=cut

sub getURL { #get a page/file using HTTP (not file:)
  my ($url, $timeout, $proxy) = @_;
  #my $debug = 8;
  my @proxy;
  my $push = 0; #don't push the timeout along if we are still receiving data
  unless ($timeout) {$timeout = 60; $push++};
  if ($proxy) {push(@proxy, $proxy)}
  else {
    @proxy = getProxies();
    unless (@proxy) {$proxy[0] = ''}; #direct
    };
  if ($debug) {print "Using proxies ", join(', ', @proxy), "\n"};
  my ($CRLF, $CR, $LF) = ("\015\012", "\015", "\012"); 
  my ($scheme, $authority, $path, $query, $fragment) =
    $url =~ 
     m|^(?:([^:/?#]+):)?(?://([^/?#]*))?([^?#]*)(?:\?([^#]*))?(?:#(.*))?|; 
  return undef unless $authority;
  #XXX - authentication with proxy not done yet.
  my $hp = $authority;
  if ($authority =~ /@/) { $hp = (split(/@/, $hp))[1]};
  my ($host, $port) = $hp =~ m|([^:]*)(?::(\d*))?|;
  unless ($port) {$port = 80};
  my $data;
  my ($cHost, $cPort);
  my $pIndex = 0;
  do {
    $proxy = $proxy[$pIndex];
    if ($debug) {print "Trying proxy $proxy\n"};
    if ($proxy) { #allow for no proxy, direct connection
      ($cHost, $cPort) = split(/:/, $proxy);
      }
    else {
      ($cHost, $cPort) = ($host, $port); #direct connection
      };
    if (my $sock = nonBlockingConnect("$cHost:$cPort", 10)) { #got connection
      my $command = "GET $url HTTP/1.0$CRLF"."Host: $host$CRLF$CRLF";
      my $sent;
      while (($sent = $sock->send($command, 0)) < length($command)) { #sent some
        substr($command, 0, $sent) = '';
        sleep(1);
        };
      shutdown($sock, 1); #finished writing
      my $sel = IO::Select->new($sock) or warn, return undef;
      my $end = time + $timeout;
      my $bit;
      while (time < $end) { #get results
        if ($sel->can_read(1)) {
          undef $!;
          my $rv = $sock->recv($bit, 1024); #peer addr or ''
          last unless (defined($rv) && (length($bit) or ($! == EAGAIN))); #peer reset
          $data .= $bit;
          $end = (time + $timeout) if $push; #XXX - experimentation
          if ($debug > 4) {
            #print "recvd: $bit\n";
            print "rv: $rv, bit: $bit, \$!: $!\n";
            };
          };
        };
      $sel->remove($sock);
      shutdown($sock, 2); #finished all
      }
    else {print "Unable to connect to $cHost:$cPort\n" if ($debug)};
    $pIndex++;
    } until ($data or !defined($proxy[$pIndex]));
  if (defined($data) and ($debug>2)) {print "Received from $url: \n$data\n"};
  unless ($data) {return undef};
  return wantarray ? split(/($CR)?$LF/, $data) : $data;
  };

=for wayne

#if ($pingInstalled and !$> and ($test==0))  #needs root
    #$p = Net::Ping->new("icmp"); # we lose the hosts which are not 
                            # icmp pingable - must change this 
                            # to tcp pings, with select to timeout
                            # the blocked ports and dead hosts quickly.
                            # icmp ping also requires root access
                            # 'coz it makes it's own packets (!), 
                            # so it won't work on shell accounts.
    #if ($p->ping($proxyAddr))  
        #&makeConnection() ;    #connect to the proxy only 
        #                   #if it's icmp reachable (for now)
    #else 
        #print STDOUT "can't ping (icmp): ignoring\n";
        #"\nhost $proxyAddr is unresolvable, dead or not icmp reachable";
        #$connected=0;
        #
    #$p->close()
    
  my $line;
  if ($^O=~/Win/) {  #windows
      $line = `ver`; $line =~ s/[\015\012]*//g;
      for ($line) {
          /Windows 95/i and do {$osVer = 1; last}; 
          /Windows 98/i and do {$osVer = 2; last}; 
          /Windows Millennium/ and do {$osVer = 2; last};
          /Windows NT Version 4.0/i and do {$osVer = 3; last}; 
          /Windows 2000/i and do {$osVer = 4; last}; 
          /Windows XP/i and do {$osVer = 4; last}; 
          #do {print "unknown o/s: $line\n"; $osVer = 2; last};
          croak "unknown o/s";
          };
      }
  elsif ($^O =~ /cygwin/) {$osVer = 10}
  else {$osVer = 10}; #incl 'netbsd'


'All ears' servers
AOL Instant Messenger:  login.oscar.aol.com
ICQ:  login.icq.com

=cut

1;


=back

=head1 SEE ALSO

See F<localProxy>, F<mergeHosts>, F<statProxy> and the other proxyTools 
(http://proxytools.sourceforge.net) for example uses of this library.  

=head1 COPYRIGHT

Copyright 1998-2001 wayne (wayne@nym.alias.net).

This library is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=cut
